-----------------------------------------------------------------------
-- Implicit queues
------------------------------------------------------------------------

{-# OPTIONS --no-termination-check #-}

-- The functions below do not pass the termination checker. The
-- functions can easily be seen to terminate, though, and the problem
-- could be addressed by indexing the queue data type on a depth
-- index.

{-# OPTIONS --no-positivity-check #-}

-- Thunk is abstract, so the positivity checker cannot "see" that it
-- is just the identity. If Thunk were inlined the Queue type would be
-- strictly positive.

module ImplicitQueue where

open import Thunk
open import Data.Product
open import Data.Nat
open import Data.Maybe
open import Data.Function

------------------------------------------------------------------------
-- Queue data type

data Queue (a : Set) : Set where
  empty    : Queue a
  single   : a -> Queue a
  twoZero  : a -> a -> Thunk 5 (Queue (a × a))      -> Queue a
  twoOne   : a -> a -> Thunk 3 (Queue (a × a)) -> a -> Queue a
  oneZero  : a ->      Thunk 2 (Queue (a × a))      -> Queue a
  oneOne   : a ->               Queue (a × a)  -> a -> Queue a

------------------------------------------------------------------------
-- Snoc

snoc : forall {a} -> Queue a -> a -> Thunk 5 (Queue a)
snoc empty                  x₁ = √ returnW 3 (single x₁)
snoc (single x₁)            x₂ = √ returnW 3
  (twoZero x₁ x₂ (returnW 4 empty))
snoc (twoZero x₁ x₂ xs₃)    x₄ = √
  pay • 2 xs₃ >>= \xs₃' -> √
  returnW 0 (twoOne x₁ x₂ xs₃' x₄)
snoc (oneZero x₁ xs₂)       x₃ = √
  xs₂ >>= \xs₂' -> √
  returnW 0 (oneOne x₁ xs₂' x₃)
snoc (twoOne x₁ x₂ xs₃ x₄)  x₅ = √
  xs₃ >>= \xs₃' -> √
  return (twoZero x₁ x₂ (snoc xs₃' (x₄ , x₅)))
snoc (oneOne x₁ xs₂ x₃)     x₄ = √
  pay • 3 (snoc xs₂ (x₃ , x₄)) >>= \xs₂₃₄ -> √
  return (oneZero x₁ xs₂₃₄)

------------------------------------------------------------------------
-- Left view data type

data Viewˡ (a : Set) : Set where
  nilˡ   : Viewˡ a
  consˡ  : a -> Thunk 4 (Queue a) -> Viewˡ a

------------------------------------------------------------------------
-- Left view

viewL : forall {a} -> Queue a -> Thunk 1 (Viewˡ a)
viewL empty                = √ return nilˡ
viewL (single x₁)          = √ return (consˡ x₁ (returnW 3 empty))
viewL (twoZero x₁ x₂ xs₃)  = √ return (consˡ x₁ (
  pay • 3 xs₃ >>= \xs₃' -> √
  return (oneZero x₂ xs₃')))
viewL {a} (oneZero x₁ xs₂) = √ return
  (consˡ x₁ (expand =<< viewL =<< xs₂))
  where
  expand : Viewˡ (a × a) -> Thunk 1 (Queue a)
  expand nilˡ                   = √ return empty
  expand (consˡ (y₁ , y₂) ys₃)  = √ return (twoZero y₁ y₂ (wait 0 ys₃))
viewL (twoOne x₁ x₂ xs₃ x₄) = √ return (consˡ x₁ (
  xs₃ >>= \xs₃' -> √
  return (oneOne x₂ xs₃' x₄)))
viewL {a} (oneOne x₁ xs₂ x₃) = √ return
  (consˡ x₁ (expand =<< viewL xs₂))
  where
  expand : Viewˡ (a × a) -> Thunk 3 (Queue a)
  expand nilˡ                   = √ returnW 1 (single x₃)
  expand (consˡ (y₁ , y₂) ys₃)  = √
    pay • 1 ys₃ >>= \ys₃' -> √
    return (twoOne y₁ y₂ ys₃' x₃)

------------------------------------------------------------------------
-- Explanation of an analysis leading up to the time bounds and
-- invariants above

{-

If you have somehow already arrived at the general structure of the
code above, and believe that all time bounds are constants, and that
all invariants are also constants, then this comment explains how to
find the exact time bounds and invariants above.

Assume that _all_ subqueues are thunked, and denote the number of
debits on the twoZero thunk by d₂₀, and so on. Also denote the number
of steps needed to snoc an element by ss, the number of steps needed
to view the head by hs, and the number of steps needed to view the
tail by ts:

  twoZero  : a -> a -> Thunk d₂₀ (Queue (a × a))      -> Queue a
  twoOne   : a -> a -> Thunk d₂₁ (Queue (a × a)) -> a -> Queue a
  oneZero  : a ->      Thunk d₁₀ (Queue (a × a))      -> Queue a
  oneOne   : a ->      Thunk d₁₁ (Queue (a × a)) -> a -> Queue a

  snoc  : forall {a} -> Queue a -> a -> Thunk ss (Queue a)
  consˡ : a -> Thunk ts (Queue a) -> Viewˡ a
  viewL : forall {a} -> Queue a -> Thunk hs (Viewˡ a)

Then the following inequalities can be derived from the basic
structure of the code (assuming a worst case scenario for how many pay
annotations have to be inserted etc.):

1                    ≤ ss
2 + (d₂₀ ∸ d₂₁)      ≤ ss
2 + (d₁₀ ∸ d₁₁)      ≤ ss
3 + d₂₁ + (ss ∸ d₂₀) ≤ ss
3 + d₁₁ + (ss ∸ d₁₀) ≤ ss

1 ≤ hs

1 + (d₂₀ ∸ d₁₀)           ≤ ts
d₁₀ + hs + 2 + (ts ∸ d₂₀) ≤ ts
1 + (d₂₁ ∸ d₁₁)           ≤ ts
d₁₁ + hs + 2 + (ts ∸ d₂₁) ≤ ts

It is immediately clear that hs = 1 is optimal. We get the following
simplified inequalities:

2 + (d₂₀ ∸ d₂₁) ≤ ss
2 + (d₁₀ ∸ d₁₁) ≤ ss
1 + (d₂₀ ∸ d₁₀) ≤ ts
1 + (d₂₁ ∸ d₁₁) ≤ ts
3 + d₂₁ + (ss ∸ d₂₀) ≤ ss
3 + d₁₁ + (ss ∸ d₁₀) ≤ ss
3 + d₁₀ + (ts ∸ d₂₀) ≤ ts
3 + d₁₁ + (ts ∸ d₂₁) ≤ ts

The last four equations are interesting, since they ensure that
setting all d:s to 0 is not OK. The following inequalities follow from
those above:

3 ≤ d₂₀ ∸ d₂₁ ≤ ss ∸ 2 ≥ d₂₁ + 1
3 ≤ d₁₀ ∸ d₁₁ ≤ ss ∸ 2 ≥ d₁₁ + 1
3 ≤ d₂₀ ∸ d₁₀ ≤ ts ∸ 1 ≥ d₁₀ + 2
3 ≤ d₂₁ ∸ d₁₁ ≤ ts ∸ 1 ≥ d₁₁ + 2

Hmm... Maybe ss = 5 can lead to a solution? No, since d₂₁ ≥ 3. So what
about ss = 6 and, continuing greedily, d₂₁ = 3, d₁₁ = 0, d₂₀ = 6? We
get:

3 ≤ d₁₀ ≤ 4
3 ≤ 6 ∸ d₁₀ ≤ ts ∸ 1 ≥ d₁₀ + 2
3 ≤ ts ∸ 1

With d₁₀ = 3 this simplifies to

5 ≤ ts ∸ 1

so we can set ts = 6. In summary:

d₁₀ = 3
d₁₁ = 0
d₂₀ = 6
d₂₁ = 3
ss  = 6
hs  = 1
ts  = 6

These values also satisfy the original inequalities above. Now, we
could be happy with these figures, but when using them it turns out
that some of the pays that we assumed are unnecessary, along with some
ticks. We can start over with the following inequalities:

1                    ≤ ss
2 + (d₂₀ ∸ d₂₁)      ≤ ss
2 + (d₁₀ ∸ d₁₁)      ≤ ss
2 + d₂₁ + (ss ∸ d₂₀) ≤ ss
2 + d₁₁ + (ss ∸ d₁₀) ≤ ss

1 ≤ hs

1 + (d₂₀ ∸ d₁₀)           ≤ ts
d₁₀ + hs + 1 + (ts ∸ d₂₀) ≤ ts
1 + (d₂₁ ∸ d₁₁)           ≤ ts
d₁₁ + hs + 2 + (ts ∸ d₂₁) ≤ ts

Solving them as above we get the following, slightly tighter bounds:

d₁₀ = 2
d₁₁ = 0
d₂₀ = 5
d₂₁ = 3
ss  = 5
hs  = 1
ts  = 4

(I do not think that there is any point in squeezing the bounds like
this, since they were already constants. However, since this example
will go in the paper I want to avoid getting questions about
suboptimal bounds.)

-}
