------------------------------------------------------------------------
-- Some monolithic vector operations
------------------------------------------------------------------------

module MonolithicVec where

open import Data.Product
open import Data.Nat
open import Data.Nat.Properties
open ℕ-semiringSolver
open import Data.Fin
open import Data.Vec hiding (reverse; _>>=_)
open import Relation.Binary.PropositionalEquality
open import Thunk
import Data.List as L
open L using ([_])
open import Algebra
import Algebra.Operations as Ops
open Ops (CommutativeSemiring.semiring ℕ-commutativeSemiring)
  using (_^_)

------------------------------------------------------------------------
-- Some boring lemmas

abstract
 private
  vecCast
    :  forall {a m₁ n₁ m₂ n₂}
    -> m₁ ≡ m₂ -> n₁ ≡ n₂
    -> Thunk m₁ (Vec a n₁)
    -> Thunk m₂ (Vec a n₂)
  vecCast ≡-refl ≡-refl x = x

  -- It would be very convenient if the proofs below, which are
  -- actually done automatically, could also be invoked
  -- automatically...

  lem₁ : forall m -> (1 + 2 * m) + 1 ≡ 2 * (1 + m)
  lem₁ m = prove (m ∷ [])
                 ((con 1 :+ con 2 :* M) :+ con 1)
                 (con 2 :* (con 1 :+ M))
                 ≡-refl
    where M = var zero

  lem₃ : forall m -> m + 0 ≡ m
  lem₃ m = prove (m ∷ []) (M :+ con 0) M ≡-refl
    where M = var zero

  lem₄ : forall m n -> m + (1 + n) ≡ (1 + m) + n
  lem₄ m n = prove (m ∷ n ∷ [])
                   (M :+ (con 1 :+ N))
                   ((con 1 :+ M) :+ N)
                   ≡-refl
    where M = var zero; N = var (suc zero)

  lem₅ : forall m n -> m + (2 + (n * (3 + m) + 1)) ≡
                       (1 + n) * (3 + m)
  lem₅ m n = prove (m ∷ n ∷ [])
                   (M :+ (con 2 :+ (N :* (con 3 :+ M) :+ con 1)))
                   ((con 1 :+ N) :* (con 3 :+ M))
                   ≡-refl
    where M = var zero; N = var (suc zero)

  lem₆ : forall n -> 1 + (2 * n + 1) ≡ 2 * (1 + n)
  lem₆ n = prove (n ∷ [])
                 (con 1 :+ (con 2 :* N :+ con 1))
                 (con 2 :* (con 1 :+ N))
                 ≡-refl
    where N = var zero

  lem₇ :  forall m n
       -> (1 + 2 * n) + (1 + n * (3 + m)) ≡ 2 + n * (5 + m)
  lem₇ m n = prove (m ∷ n ∷ [])
                   ((con 1 :+ con 2 :* N) :+
                    (con 1 :+ N :* (con 3 :+ M)))
                   (con 2 :+ N :* (con 5 :+ M))
                   ≡-refl
    where M = var zero; N = var (suc zero)

  -- The automatic solver cannot handle ∸, and I cannot be bothered to
  -- prove this formally.
  postulate
    lem₈ : forall n
         -> 1 + (((6 + 2 * n) * 2 ^ n ∸ 5) +
                 ((1 + (((6 + 2 * n) * 2 ^ n ∸ 5) + (1 + 2 * 2 ^ n))) +
                  (2 + 2 * 2 ^ n)))
            ≡
            (6 + 2 * suc n) * 2 ^ suc n ∸ 5

  lem₈' : forall n -> 2 ^ n + 2 ^ n ≡ 2 * 2 ^ n
  lem₈' n = prove (2 ^ n ∷ []) (N :+ N) (con 2 :* N) ≡-refl
    where N = var zero

  lem₉ :  forall m n
       -> 1 + n * (3 + 2 * m) + (2 + 2 * m) ≡
          (1 + n) * (3 + 2 * m)
  lem₉ m n = prove (m ∷ n ∷ [])
                   (con 1 :+ N :* (con 3 :+ con 2 :* M)
                          :+ (con 2 :+ con 2 :* M))
                   ((con 1 :+ N) :* (con 3 :+ con 2 :* M))
                   ≡-refl
    where M = var zero; N = var (suc zero)

  lem₁₁ : forall m -> 1 + (2 * m + 1) ≡ 2 * (1 + m)
  lem₁₁ m = prove (m ∷ [])
                  (con 1 :+ (con 2 :* M :+ con 1))
                  (con 2 :* (con 1 :+ M))
                  ≡-refl
    where M = var zero

  lem₁₃ :  forall n k
        -> 1 + (2 * k + (2 + (n * (4 + 2 * k) + 1))) ≡
           (1 + n) * (4 + 2 * k)
  lem₁₃ n k = prove
      (n ∷ k ∷ [])
      (con 1 :+ (con 2 :* K :+ ((con 2 :+ N :* X) :+ con 1)))
      ((con 1 :+ N) :* X)
      ≡-refl
    where N = var zero; K = var (suc zero); X = con 4 :+ con 2 :* K

------------------------------------------------------------------------
-- More interesting functions

head√ : forall {a n} -> Vec a (1 + n) -> Thunk 1 a
head√ (x ∷ xs) = √ return x

tail√ : forall {a n} -> Vec a (1 + n) -> Thunk 1 (Vec a n)
tail√ (x ∷ xs) = √ return xs

infixr 5 _++√_

_++√_ :  forall {a m n}
      -> Vec a m -> Vec a n -> Thunk (1 + 2 * m) (Vec a (m + n))
[]                 ++√ ys = √ return ys
(_∷_ {n = m} x xs) ++√ ys = √
  cast (lem₁ m)
    (xs ++√ ys >>= \xsys -> √
     return (x ∷ xsys))

reverse : forall {a m} -> Vec a m -> Thunk (2 + m) (Vec a m)
reverse xs = √
  vecCast ≡-refl (lem₃ _)
    (rev xs [])
  where
  rev :  forall {a m n}
      -> Vec a m -> Vec a n -> Thunk (1 + m) (Vec a (m + n))
  rev []                 ys = √ return ys
  rev (_∷_ {n = m} x xs) ys = √
    vecCast ≡-refl (lem₄ m _)
      (rev xs (x ∷ ys))

forceVec :  forall {m n a}
         -> Vec (Thunk m a) n -> Thunk (1 + n * (3 + m)) (Vec a n)
forceVec             []       = √ return []
forceVec {m} {suc n} (x ∷ xs) = √
  cast (lem₅ m n)
  (x >>= \x' -> √
   forceVec xs >>= \xs' -> √
   return (x' ∷ xs'))

map√ :  forall {n a b}
     -> (a -> b) -> Vec a n -> Thunk (1 + 2 * n) (Vec b n)
map√         f []       = √ return []
map√ {suc n} f (x ∷ xs) = √
  cast (lem₆ n)
  (map√ f xs >>= \xs' -> √
   return (f x ∷ xs'))

map√√ :  forall {m n a b}
      -> (a -> Thunk m b) -> Vec a n
      -> Thunk (2 + n * (5 + m)) (Vec b n)
map√√ {m} {n} f xs =
  cast (lem₇ m n)
  (forceVec =<< map√ f xs)

-- The powerset function. Note that the recursive calls are not
-- shared; this is sometimes an advantage.

℘ :  forall {a n}
  -> Vec a n -> Thunk ((6 + 2 * n) * 2 ^ n ∸ 5) (Vec [ a ] (2 ^ n))
℘             []       = √ return (L.[] ∷ [])
℘ {n = suc n} (x ∷ xs) = vecCast (lem₈ n) (lem₈' n) (√
   ℘ xs                       >>= \ ℘xs  -> √
   (map√ (L._∷_ x) =<< ℘ xs)  >>= \x∷℘xs -> √
   x∷℘xs ++√ ℘xs)

concat√ : forall {a m n}
       -> Vec (Vec a m) n -> Thunk (1 + n * (3 + 2 * m)) (Vec a (n * m))
concat√         []                   = √ return []
concat√ {m = m} (_∷_ {n = n} xs xss) = √
  cast (lem₉ m n )
    (concat√ xss >>= \xs' -> √
     xs ++√ xs')

splitAt√ :  forall {a} m {n}
         -> Vec a (m + n) -> Thunk (1 + 2 * m) (Vec a m × Vec a n)
splitAt√         zero            xs       = √ return ([] , xs)
splitAt√ {a = a} (suc m) {n = n} (x ∷ xs) = √
  cast (lem₁₁ m) (helper =<< splitAt√ m xs)
  where
  helper : Vec a m × Vec a n -> Thunk 1 (Vec a (1 + m) × Vec a n)
  helper (ys , zs) = √ return (x ∷ ys , zs)

group√ :  forall {a n}
       -> (k : ℕ) -> Vec a (n * k)
       -> Thunk (1 + n * (4 + 2 * k)) (Vec (Vec a k) n)
group√         {n = zero}  k [] = √ return []
group√ {a = a} {n = suc n} k xs = √
  cast (lem₁₃ n k)
    (helper =<< splitAt√ k xs)
  where
  helper :  Vec a k × Vec a (n * k)
         -> Thunk (1 + (1 + n * (4 + 2 * k)) + 1)
                  (Vec (Vec a k) (1 + n))
  helper (init , rest) = √
    group√ {n = n} k rest >>= \grouped -> √
    return (init ∷ grouped)
