module Thunk where

open import Data.Nat
open import Data.Product
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning
open import Algebra
open import Algebra.Structures
open import Data.Function
open import Data.Nat.Properties

open import Thunk.Context public

infix  5 _[_]

infixl 1 _>>=_
infixr 1 _=<<_
infix  0 √_

------------------------------------------------------------------------
-- Core functionality (except for pay)

abstract

  -- A value of type Thunk n₁ (Thunk n₂ … (Thunk n_k τ)…), where τ is
  -- not Thunk something, takes Σ n_i steps to evaluate to weak head
  -- normal form (assuming a total language).

  Thunk : ℕ -> Set -> Set
  Thunk n a = a

  -- Forces computations; not run-time.

  force : forall {n a} -> Thunk n a -> a
  force x = x

  -- Combines computations.

  _>>=_
    :  forall {a b m n}
    -> Thunk m a -> (a -> Thunk n b) -> Thunk (m + n) b
  x >>= f = f x

  -- Injects thunk-normal forms into the thunk monad.

  return : forall {a} -> a -> Thunk 0 a
  return x = x

  -- A step or tick; represents one unit of computation.

  √_ : forall {a n} -> Thunk n a -> Thunk (1 + n) a
  √_ x = x

-- Could be a derived operation (implemented using √), but with worse
-- time bounds.

_=<<_ :  forall {a b m n}
      -> (a -> Thunk m b) -> Thunk n a -> Thunk (n + m) b
f =<< c = c >>= f

------------------------------------------------------------------------
-- Paying off debits

_[_] : Ctxt -> Set -> Set
_[_] = fill Thunk

abstract

  -- Pays off debits, possibly deep inside a data structure; a context
  -- is used to describe the position being paid off.

  pay :  forall {a n}
      -> (C : Ctxt) -> (m : ℕ)
      -> C [ Thunk n a ] -> Thunk m (C [ Thunk (n ∸ m) a ])
  pay _ _ x = x

-- Contexts have to be quite limited to avoid cheating. Consider the
-- following cases:

-- * Multiple-hole contexts, like • × •:
--   Two or more holes could be paid off at the price of one.

-- * Contexts for more complicated data structure, like Vec n •:
--   All elements in the vector could be paid off at the price of only
--   one.

-- * Contexts involving the codomain of a function, like ℕ -> •:
--   All uses of the function could be made cheaper by paying once.

-- * Contexts involving the domain of a function, like • -> Thunk n ℕ:
--   This does not seem to be dangerous, but also does not seem to
--   help.

-- The definition of contexts could perhaps be more general,
-- especially if Agda had support for datatype-generic programming.
-- The definition above suffices for my examples, though.

------------------------------------------------------------------------
-- Utilities

-- Used to insert equality proofs, if necessary. cast is assumed to
-- vanish completely when the program is compiled (if it is used fully
-- applied).

cast : forall {a n₁ n₂} -> n₁ ≡ n₂ -> Thunk n₁ a -> Thunk n₂ a
cast {a} = ≡-subst (\n -> Thunk n a)

-- Increases the step count for a computation.
--
-- It would be very convenient if the "1 +" in the type of wait' could
-- be omitted... (It can, if wait' is taken as a primitive. I want the
-- core library to be as simple as possible, though. Update: In order
-- to analyse generalised pay I have included a generalised version of
-- wait, without the "1 +", in the formalisation. I may not have time
-- to include it in the library, though.)

private

  wait' : forall {a m} n -> Thunk m a -> Thunk (1 + n + m) a
  wait' zero    x = √ x
  wait' (suc n) x = √ wait' n x

-- Since wait' can be implemented with the given type signature it is
-- safe to add a more efficient variant to the library core:

abstract

  wait : forall {a m} n -> Thunk m a -> Thunk (1 + n + m) a
  wait _ x = x

wait-≤ : forall {a m n} -> m ≤ n -> Thunk m a -> Thunk (2 + n) a
wait-≤ {m = m} {n = n} m≤n x = √ cast lem (wait (n ∸ m) x)
  where
  abstract
    open IsCommutativeSemiring _
           (CommutativeSemiring.isCommutativeSemiring
              ℕ-commutativeSemiring)

    lem : 1 + (n ∸ m) + m ≡ 1 + n
    lem =                 begin
          1 + (n ∸ m) + m
                          ≡⟨ ≡-cong suc $ +-comm (n ∸ m) m ⟩
          1 + m + (n ∸ m)
                          ≡⟨ ≡-cong suc $ m+n∸m≡n m≤n ⟩
          1 + n
                          ∎

-- Suboptimal variant of return:

private

  returnW' : forall {a} n -> a -> Thunk (1 + n) a
  returnW' zero    x = √ return x
  returnW' (suc n) x = √ returnW' n x

-- Optimised variant:

abstract

  returnW : forall {a} n -> a -> Thunk (1 + n) a
  returnW _ x = x

------------------------------------------------------------------------
-- Facilities for reasoning about thunks

-- The functions below are all non-run-time.

-- Note that I have not verified that the functions below are safe to
-- use, even in a compile-time setting. For instance, if they made it
-- possible to prove that Thunk n a = a, then it would be easy to
-- cheat. I do believe that the functions are safe, but I have not
-- verified it. These functions are not used by the case studies
-- mentioned in the paper.

import Thunk.Concrete as C

abstract

  -- Converts abstract (run-time) thunks to concrete thunks.

  ⟦_⟧ : forall {n a} -> Thunk n a -> C.Thunk n a
  ⟦_⟧ {n} x = C.return⋆ n x

  ⟦_∣_⟧ : forall {n a} C -> C [ Thunk n a ] -> C._[_] C (C.Thunk n a)
  ⟦ •          ∣ x     ⟧ = ⟦ x ⟧
  ⟦ const• a   ∣ x     ⟧ = x
  ⟦ Thunk• n C ∣ x     ⟧ = ⟦ ⟦ C ∣ x ⟧ ⟧
  ⟦ C •× a     ∣ x , y ⟧ = ⟦ C ∣ x ⟧ , y
  ⟦ a ×• C     ∣ x , y ⟧ = x , ⟦ C ∣ y ⟧

  -- Converts concrete thunks to abstract thunks.

  ⟪_⟫ : forall {n a} -> C.Thunk n a -> Thunk n a
  ⟪ x ⟫ = C.force x

  -- Various properties relating the abstract and concrete
  -- implementations.

  prop-⟪⟫-⟦⟧ :  forall {n a} (x : Thunk n a)
             -> ⟪ ⟦ x ⟧ ⟫ ≡ x
  prop-⟪⟫-⟦⟧ {n} x = C.prop-force-return⋆ n x

  prop-⟦⟧-⟪⟫ :  forall {n a} (x : C.Thunk n a)
             -> ⟦ ⟪ x ⟫ ⟧ ≡ x
  prop-⟦⟧-⟪⟫ x = C.prop-return⋆-force x

  prop-force :  forall {n a} (x : Thunk n a)
             -> force x ≡ C.force ⟦ x ⟧
  prop-force {n} x = ≡-sym $ C.prop-force-return⋆ n x

  prop-return :  forall {a} (x : a)
              -> ⟦ return x ⟧ ≡ C.return x
  prop-return x = byDef

  prop-√ :  forall {n a} (x : Thunk n a)
         -> ⟦ √ x ⟧ ≡ C.√_ ⟦ x ⟧
  prop-√ x = byDef

  prop->>= :  forall {n₁ n₂ a b} (x : Thunk n₁ a) (f : a -> Thunk n₂ b)
           -> ⟦ x >>= f ⟧ ≡ (⟦ x ⟧ ⟨ C._>>=_ ⟩ \y -> ⟦ f y ⟧)
  prop->>= {zero}  x f = byDef
  prop->>= {suc n} x f = ≡-cong C.√_ (prop->>= {n} x f)

  prop-pay
    :  forall {a} C n₂ {n₁} (x : C [ Thunk n₁ a ])
    -> C.map (\y -> ⟦ C ∣ y ⟧) ⟦ pay C n₂ x ⟧ ≡ C.pay C n₂ ⟦ C ∣ x ⟧
  prop-pay • zero              x = byDef
  prop-pay • (suc n₂) {zero}   x =
    ≡-cong C.√_ $ C.prop-map-return⋆ n₂ (⟦_∣_⟧ •) x
  prop-pay • (suc n₂) {suc n₁} x =
    ≡-cong C.√_ $ prop-pay • n₂ x
  prop-pay {a} (const• b) n₂ {n₁} x =
    C.prop-map-return⋆ n₂ (⟦_∣_⟧ {n₁} {a} (const• b)) x
  prop-pay (Thunk• n C) n₂ x =
                                                            begin
    C.map (⟦_∣_⟧ (Thunk• n C)) (C.return⋆ n₂ x)
                                                            ≡⟨ byDef ⟩
    C.map (C.return⋆ n ∘ ⟦_∣_⟧ C) (C.return⋆ n₂ x)
                                                            ≡⟨ ≡-sym $ C.map-functorial (C.return⋆ n) (⟦_∣_⟧ C)
                                                                                        (C.return⋆ n₂ x) ⟩
    C.map (C.return⋆ n) (C.map (⟦_∣_⟧ C) (C.return⋆ n₂ x))
                                                            ≡⟨ ≡-cong (C.map (C.return⋆ n)) (prop-pay C n₂ x) ⟩
    C.map (C.return⋆ n) (C.pay C n₂ ⟦ C ∣ x ⟧)
                                                            ≡⟨ ≡-sym $ C.prop-swap-map n (C.pay C n₂ ⟦ C ∣ x ⟧) ⟩
    C.swap (C.return⋆ n (C.pay C n₂ ⟦ C ∣ x ⟧))
                                                            ≡⟨ ≡-sym $ ≡-cong C.swap $
                                                                 C.prop-map-return⋆ n (C.pay C n₂) ⟦ C ∣ x ⟧ ⟩
    C.swap (C.map (C.pay C n₂) (C.return⋆ n ⟦ C ∣ x ⟧))
                                                            ∎
  prop-pay (b ×• C) n₂ (x , y) =
                                                            begin
    C.map (⟦_∣_⟧ (b ×• C)) (C.return⋆ n₂ (x , y))
                                                            ≡⟨ C.prop-map-return⋆ n₂ (⟦_∣_⟧ (b ×• C)) (x , y) ⟩
    C.return⋆ n₂ (x , ⟦ C ∣ y ⟧)
                                                            ≡⟨ ≡-sym $ C.prop-map-return⋆ n₂ (\y -> x , ⟦ C ∣ y ⟧) y ⟩
    C.map (\y -> x , ⟦ C ∣ y ⟧) (C.return⋆ n₂ y)
                                                            ≡⟨ ≡-sym $ C.map-functorial (\y -> x , y) (⟦_∣_⟧ C)
                                                                                        (C.return⋆ n₂ y) ⟩
    C.map (\y -> x , y) (C.map (⟦_∣_⟧ C) (C.return⋆ n₂ y))
                                                            ≡⟨ ≡-cong (C.map (\y -> x , y)) (prop-pay C n₂ y) ⟩
    C.map (\y -> x , y) (C.pay C n₂ ⟦ C ∣ y ⟧)
                                                            ∎
  prop-pay (C •× b) n₂ (x , y) =
                                                            begin
    C.map (⟦_∣_⟧ (C •× b)) (C.return⋆ n₂ (x , y))
                                                            ≡⟨ C.prop-map-return⋆ n₂ (⟦_∣_⟧ (C •× b)) (x , y) ⟩
    C.return⋆ n₂ (⟦ C ∣ x ⟧ , y)
                                                            ≡⟨ ≡-sym $ C.prop-map-return⋆ n₂ (\x -> ⟦ C ∣ x ⟧ , y) x ⟩
    C.map (\x -> ⟦ C ∣ x ⟧ , y) (C.return⋆ n₂ x)
                                                            ≡⟨ ≡-sym $ C.map-functorial (\x -> x , y) (⟦_∣_⟧ C)
                                                                                        (C.return⋆ n₂ x) ⟩
    C.map (\x -> x , y) (C.map (⟦_∣_⟧ C) (C.return⋆ n₂ x))
                                                            ≡⟨ ≡-cong (C.map (\x -> x , y)) (prop-pay C n₂ x) ⟩
    C.map (\x -> x , y) (C.pay C n₂ ⟦ C ∣ x ⟧)
                                                            ∎
