------------------------------------------------------------------------
-- General (not necessarily run-time) thunked terms
------------------------------------------------------------------------

-- TypeSystem defines the simple terms and the thunked, run-time
-- terms. The general thunked terms are not needed for the
-- formalisation, but included in the paper to make it easier to
-- follow. This module defines the general thunked terms, to show that
-- the definition in the paper is well-formed. It is also shown how
-- run-time terms can be translated into general terms (by making the
-- implicit ticks explicit), and how general terms can be erased.

module Thunked where

open import Kind
open import TypeSystem
open import Erasure

open import Data.Nat using (zero; suc; _+_; _∸_)
open import Relation.Binary.PropositionalEquality
  using (_≡_; ≡-refl; ≡-subst; ≡-cong; ≡-cong₂)

infixl 70 _·⁺_
infixr 60 _,⁺_
infix  50 λ⁺_ λ²⁺_
infixl 46 _>>=⁺_
infix  43 √⁺_
infix  10 _⊢⁺_

-- General thunked terms. (Note that force is included.)

data _⊢⁺_ (Γ : Ctxt thunked) : Ty thunked -> Set where
  -- Lambda calculus.
  var⁺ : forall {τ}       -> Γ ∋ τ -> Γ ⊢⁺ τ
  λ⁺_  : forall {σ τ}     -> Γ ▻ σ ⊢⁺ τ -> Γ ⊢⁺ σ → τ
  λ²⁺_ : forall {σ₁ σ₂ τ} -> Γ ▻ σ₁ ▻ σ₂ ⊢⁺ τ -> Γ ⊢⁺ σ₁ → σ₂ → τ
  _·⁺_ : forall {σ τ}     -> Γ ⊢⁺ σ → τ -> Γ ⊢⁺ σ -> Γ ⊢⁺ τ
  fix⁺ : forall {σ}       -> Γ ▻ σ ⊢⁺ σ -> Γ ⊢⁺ σ

  -- Natural numbers.
  z⁺      :  Γ ⊢⁺ Nat
  s⁺      :  Γ ⊢⁺ Nat → Nat
  natrec⁺ :  forall {σ}
          -> Γ ⊢⁺ σ -> Γ ▻ Nat ▻ σ ⊢⁺ σ -> Γ ⊢⁺ Nat → σ

  -- Products.
  _,⁺_     :  forall {σ τ}     -> Γ ⊢⁺ σ -> Γ ⊢⁺ τ -> Γ ⊢⁺ σ × τ
  uncurry⁺ :  forall {σ₁ σ₂ τ}
           -> Γ ▻ σ₁ ▻ σ₂ ⊢⁺ τ -> Γ ⊢⁺ σ₁ × σ₂ → τ

  -- Thunk library primitives.
  force⁺  :  forall {n σ} -> Γ ⊢⁺ Thunk n σ -> Γ ⊢⁺ σ
  return⁺ :  forall {σ}   -> Γ ⊢⁺ σ -> Γ ⊢⁺ Thunk zero σ
  _>>=⁺_  :  forall {σ τ n₁ n₂}
          -> Γ ⊢⁺ Thunk n₁ σ -> Γ ⊢⁺ σ → Thunk n₂ τ
          -> Γ ⊢⁺ Thunk (n₁ + n₂) τ
  √⁺_     :  forall {σ n}
          -> Γ ⊢⁺ Thunk n σ -> Γ ⊢⁺ Thunk (suc n) σ
  pay⁺    :  forall {σ n₁} n₂
          -> Γ ⊢⁺ Thunk n₁ σ
          -> Γ ⊢⁺ Thunk n₂ (Thunk (n₁ ∸ n₂) σ)
  gpay⁺   :  forall {σ n₁} C n₂
          -> Γ ⊢⁺ C [ Thunk n₁ σ ]
          -> Γ ⊢⁺ Thunk n₂ (C [ Thunk (n₁ ∸ n₂) σ ])
  gwait⁺  :  forall {σ n₁} C n₂
          -> Γ ⊢⁺ C [ Thunk n₁ σ ]
          -> Γ ⊢⁺ C [ Thunk (n₂ + n₁) σ ]

-- This definition exposes the implicit ticks present in the
-- representation of run-time terms.

tmToTm⁺ : forall {Γ τ} -> Γ ⊢ τ -> Γ ⊢⁺ τ
tmToTm⁺ (var x)        = var⁺ x
tmToTm⁺ (λ e)          = λ⁺ (√⁺ tmToTm⁺ e)
tmToTm⁺ (λ² e)         = λ²⁺ (√⁺ tmToTm⁺ e)
tmToTm⁺ (e₁ · e₂)      = tmToTm⁺ e₁ ·⁺ tmToTm⁺ e₂
tmToTm⁺ (fix e)        = fix⁺ (√⁺ tmToTm⁺ e)
tmToTm⁺ z              = z⁺
tmToTm⁺ s              = s⁺
tmToTm⁺ (natrec e₁ e₂) = natrec⁺ (√⁺ tmToTm⁺ e₁) (√⁺ tmToTm⁺ e₂)
tmToTm⁺ (e₁ , e₂)      = (tmToTm⁺ e₁ ,⁺ tmToTm⁺ e₂)
tmToTm⁺ (uncurry e)    = uncurry⁺ (√⁺ tmToTm⁺ e)
tmToTm⁺ (return e)     = return⁺ (tmToTm⁺ e)
tmToTm⁺ (e₁ >>= e₂)    = tmToTm⁺ e₁ >>=⁺ tmToTm⁺ e₂
tmToTm⁺ (√ e)          = √⁺ tmToTm⁺ e
tmToTm⁺ (pay n₂ e)     = pay⁺ n₂ (tmToTm⁺ e)
tmToTm⁺ (gpay C n₂ e)  = gpay⁺ C n₂ (tmToTm⁺ e)
tmToTm⁺ (gwait C n₂ e) = gwait⁺ C n₂ (tmToTm⁺ e)

-- This definition shows that the definition of erasure used in the
-- paper is well-typed.

⌜_⌝⁺ : forall {Γ τ} -> Γ ⊢⁺ τ -> ⌜ Γ ⌝▻ ⊢ ⌜ τ ⌝⋆
⌜ var⁺ x ⌝⁺                  = var ⌜ x ⌝∋
⌜ λ⁺ t ⌝⁺                    = λ ⌜ t ⌝⁺
⌜ λ²⁺ t ⌝⁺                   = λ² ⌜ t ⌝⁺
⌜ t₁ ·⁺ t₂ ⌝⁺                = ⌜ t₁ ⌝⁺ · ⌜ t₂ ⌝⁺
⌜ fix⁺ t ⌝⁺                  = fix ⌜ t ⌝⁺
⌜ z⁺ ⌝⁺                      = z
⌜ s⁺ ⌝⁺                      = s
⌜ natrec⁺ t₁ t₂ ⌝⁺           = natrec ⌜ t₁ ⌝⁺ ⌜ t₂ ⌝⁺
⌜ (t₁ ,⁺ t₂) ⌝⁺              = (⌜ t₁ ⌝⁺ , ⌜ t₂ ⌝⁺)
⌜ uncurry⁺ t ⌝⁺              = uncurry ⌜ t ⌝⁺
⌜ force⁺ t ⌝⁺                = ⌜ t ⌝⁺
⌜ return⁺ t ⌝⁺               = ⌜ t ⌝⁺
⌜ t₁ >>=⁺ t₂ ⌝⁺              = ⌜ t₂ ⌝⁺ · ⌜ t₁ ⌝⁺
⌜ √⁺ t ⌝⁺                    = ⌜ t ⌝⁺
⌜ pay⁺ n₂ t ⌝⁺               = ⌜ t ⌝⁺
⌜_⌝⁺ {Γ = Γ} (gpay⁺ C n₂ t)  =
  ≡-subst (\τ -> ⌜ Γ ⌝▻ ⊢ τ) (erase-lemma C) ⌜ t ⌝⁺
⌜_⌝⁺ {Γ = Γ} (gwait⁺ C n₂ t) =
  ≡-subst (\τ -> ⌜ Γ ⌝▻ ⊢ τ) (erase-lemma C) ⌜ t ⌝⁺

-- The definition of erasure actually used is equivalent to the
-- composition of the definitions above.

⌜_⌝-lemma : forall {Γ τ} (e : Γ ⊢ τ) -> ⌜ e ⌝ ≡ ⌜ tmToTm⁺ e ⌝⁺
⌜_⌝-lemma (var x)        = ≡-refl
⌜_⌝-lemma (λ t)          = ≡-cong λ_ (⌜_⌝-lemma t)
⌜_⌝-lemma (λ² t)         = ≡-cong λ²_ (⌜_⌝-lemma t)
⌜_⌝-lemma (t₁ · t₂)      = ≡-cong₂ _·_ (⌜_⌝-lemma t₁) (⌜_⌝-lemma t₂)
⌜_⌝-lemma (fix t)        = ≡-cong fix (⌜_⌝-lemma t)
⌜_⌝-lemma z              = ≡-refl
⌜_⌝-lemma s              = ≡-refl
⌜_⌝-lemma (natrec t₁ t₂) = ≡-cong₂ natrec (⌜_⌝-lemma t₁) (⌜_⌝-lemma t₂)
⌜_⌝-lemma (t₁ , t₂)      = ≡-cong₂ _,_ (⌜_⌝-lemma t₁) (⌜_⌝-lemma t₂)
⌜_⌝-lemma (uncurry t)    = ≡-cong uncurry (⌜_⌝-lemma t)
⌜_⌝-lemma (return t)     = ⌜_⌝-lemma t
⌜_⌝-lemma (t₁ >>= t₂)    = ≡-cong₂ _·_ (⌜_⌝-lemma t₂) (⌜_⌝-lemma t₁)
⌜_⌝-lemma (√ t)          = ⌜_⌝-lemma t
⌜_⌝-lemma (pay n₂ t)     = ⌜_⌝-lemma t
⌜_⌝-lemma (gpay {σ = σ} {n₁ = n₁} C n₂ t)
  with ⌜ C [ Thunk (n₁ ∸ n₂) σ ] ⌝⋆
     | erase-lemma {m = n₁} {n = n₁ ∸ n₂} {τ = σ} C
... | ._ | ≡-refl = ⌜_⌝-lemma t
⌜_⌝-lemma (gwait {σ = σ} {n₁ = n₁} C n₂ t)
  with ⌜ C [ Thunk (n₂ + n₁) σ ] ⌝⋆
     | erase-lemma {m = n₁} {n = n₂ + n₁} {τ = σ} C
... | ._ | ≡-refl = ⌜_⌝-lemma t
